@@ -31,6 +31,7 @@
  */
 
 /*** MODULEINFO
+	<depend>res_curl</depend>
 	<depend>curl</depend>
 	<support_level>core</support_level>
  ***/
@@ -52,6 +53,9 @@
 
 /*** DOCUMENTATION
 	<function name="CURL" language="en_US">
+		<since>
+			<version>10.0.0</version>
+		</since>
 		<synopsis>
 			Retrieve content from a remote web or ftp server
 		</synopsis>
@@ -94,6 +98,9 @@
 		</see-also>
 	</function>
 	<function name="CURLOPT" language="en_US">
+		<since>
+			<version>10.0.0</version>
+		</since>
 		<synopsis>
 			Sets various options for future invocations of CURL.
 		</synopsis>
@@ -110,6 +117,9 @@
 					<enum name="dnstimeout">
 						<para>Number of seconds to wait for DNS to be resolved</para>
 					</enum>
+					<enum name="followlocation">
+						<para>Whether or not to follow HTTP 3xx redirects (boolean)</para>
+					</enum>
 					<enum name="ftptext">
 						<para>For FTP URIs, force a text transfer (boolean)</para>
 					</enum>
@@ -121,12 +131,19 @@
 						<para>Include header information in the result
 						(boolean)</para>
 					</enum>
+					<enum name="httpheader">
+						<para>Add HTTP header. Multiple calls add multiple headers.
+						Setting of any header will remove the default
+						"Content-Type application/x-www-form-urlencoded"</para>
+					</enum>
 					<enum name="httptimeout">
 						<para>For HTTP(S) URIs, number of seconds to wait for a
 						server response</para>
 					</enum>
 					<enum name="maxredirs">
-						<para>Maximum number of redirects to follow</para>
+						<para>Maximum number of redirects to follow. The default is -1,
+						which allows for unlimited redirects. This only makes sense when
+						followlocation is also set.</para>
 					</enum>
 					<enum name="proxy">
 						<para>Hostname or IP address to use as a proxy server</para>
@@ -162,6 +179,46 @@
 						<para>Whether to verify the server certificate against
 						a list of known root certificate authorities (boolean).</para>
 					</enum>
+					<enum name="ssl_verifyhost">
+						<para>Whether to verify the host in the server's TLS certificate.
+						Set to 2 to verify the host, 0 to ignore the host.</para>
+					</enum>
+					<enum name="ssl_cainfo">
+						<para>Path to a file holding one or more certificates to verify
+						the peer's certificate with. Only used when <literal>ssl_verifypeer</literal>
+						is enabled.</para>
+					</enum>
+					<enum name="ssl_capath">
+						<para>Path to a directory holding multiple CA certificates to
+						verify the peer's certificate with. Only used when <literal>ssl_verifypeer</literal>
+						is enabled.</para>
+					</enum>
+					<enum name="ssl_cert">
+						<para>Path to a file containing a client certificate. Default format
+						is PEM, and can be changed with <literal>ssl_certtype</literal>.</para>
+					</enum>
+					<enum name="ssl_certtype">
+						<para>The format of the <literal>ssl_cert</literal> file.</para>
+						<enumlist>
+							<enum name="PEM" />
+							<enum name="DER" />
+						</enumlist>
+					</enum>
+					<enum name="ssl_key">
+						<para>Path to a file containing a client private key. Default format
+						is PEM, and can be changed with <literal>ssl_keytype</literal></para>
+					</enum>
+					<enum name="ssl_keytype">
+						<para>The format of the <literal>ssl_key</literal> file.</para>
+						<enumlist>
+							<enum name="PEM" />
+							<enum name="DER" />
+							<enum name="ENG" />
+						</enumlist>
+					</enum>
+					<enum name="ssl_keypasswd">
+						<para>The passphrase to use the <literal>ssl_key</literal> file.</para>
+					</enum>
 					<enum name="hashcompat">
 						<para>Assuming the responses will be in <literal>key1=value1&amp;key2=value2</literal>
 						format, reformat the response such that it can be used
@@ -176,12 +233,15 @@
 							</enum>
 						</enumlist>
 					</enum>
+					<enum name="failurecodes">
+						<para>A comma separated list of HTTP response codes to be treated as errors</para>
+					</enum>
 				</enumlist>
 			</parameter>
 		</syntax>
 		<description>
 			<para>Options may be set globally or per channel.  Per-channel
-			settings will override global settings.</para>
+			settings will override global settings. Only HTTP headers are added instead of overriding</para>
 		</description>
 		<see-also>
 			<ref type="function">CURL</ref>
@@ -195,6 +255,8 @@
 
 #define CURLOPT_SPECIAL_HASHCOMPAT ((CURLoption) -500)
 
+#define CURLOPT_SPECIAL_FAILURE_CODE 999
+
 static void curlds_free(void *data);
 
 static const struct ast_datastore_info curl_info = {
@@ -243,6 +305,9 @@ static int parse_curlopt_key(const char *name, CURLoption *key, enum optiontype
 	if (!strcasecmp(name, "header")) {
 		*key = CURLOPT_HEADER;
 		*ot = OT_BOOLEAN;
+	} else if (!strcasecmp(name, "httpheader")) {
+		*key = CURLOPT_HTTPHEADER;
+		*ot = OT_STRING;
 	} else if (!strcasecmp(name, "proxy")) {
 		*key = CURLOPT_PROXY;
 		*ot = OT_STRING;
@@ -261,6 +326,9 @@ static int parse_curlopt_key(const char *name, CURLoption *key, enum optiontype
 	} else if (!strcasecmp(name, "proxyuserpwd")) {
 		*key = CURLOPT_PROXYUSERPWD;
 		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "followlocation")) {
+		*key = CURLOPT_FOLLOWLOCATION;
+		*ot = OT_BOOLEAN;
 	} else if (!strcasecmp(name, "maxredirs")) {
 		*key = CURLOPT_MAXREDIRS;
 		*ot = OT_INTEGER;
@@ -298,9 +366,36 @@ static int parse_curlopt_key(const char *name, CURLoption *key, enum optiontype
 	} else if (!strcasecmp(name, "ssl_verifypeer")) {
 		*key = CURLOPT_SSL_VERIFYPEER;
 		*ot = OT_BOOLEAN;
+	} else if (!strcasecmp(name, "ssl_verifyhost")) {
+		*key = CURLOPT_SSL_VERIFYHOST;
+		*ot = OT_INTEGER;
+	} else if (!strcasecmp(name, "ssl_cainfo")) {
+		*key = CURLOPT_CAINFO;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_capath")) {
+		*key = CURLOPT_CAPATH;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_cert")) {
+		*key = CURLOPT_SSLCERT;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_certtype")) {
+		*key = CURLOPT_SSLCERTTYPE;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_key")) {
+		*key = CURLOPT_SSLKEY;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_keytype")) {
+		*key = CURLOPT_SSLKEYTYPE;
+		*ot = OT_STRING;
+	} else if (!strcasecmp(name, "ssl_keypasswd")) {
+		*key = CURLOPT_KEYPASSWD;
+		*ot = OT_STRING;
 	} else if (!strcasecmp(name, "hashcompat")) {
 		*key = CURLOPT_SPECIAL_HASHCOMPAT;
 		*ot = OT_ENUM;
+	} else if (!strcasecmp(name, "failurecodes")) {
+		*key = CURLOPT_SPECIAL_FAILURE_CODE;
+		*ot = OT_STRING;
 	} else {
 		return -1;
 	}
@@ -316,16 +411,19 @@ static int acf_curlopt_write(struct ast_channel *chan, const char *cmd, char *na
 	enum optiontype ot;
 
 	if (chan) {
+		ast_channel_lock(chan);
 		if (!(store = ast_channel_datastore_find(chan, &curl_info, NULL))) {
 			/* Create a new datastore */
 			if (!(store = ast_datastore_alloc(&curl_info, NULL))) {
 				ast_log(LOG_ERROR, "Unable to allocate new datastore.  Cannot set any CURL options\n");
+				ast_channel_unlock(chan);
 				return -1;
 			}
 
 			if (!(list = ast_calloc(1, sizeof(*list)))) {
 				ast_log(LOG_ERROR, "Unable to allocate list head.  Cannot set any CURL options\n");
 				ast_datastore_free(store);
+				ast_channel_unlock(chan);
 				return -1;
 			}
 
@@ -335,6 +433,7 @@ static int acf_curlopt_write(struct ast_channel *chan, const char *cmd, char *na
 		} else {
 			list = store->data;
 		}
+		ast_channel_unlock(chan);
 	} else {
 		/* Populate the global structure */
 		list = &global_curl_info;
@@ -412,16 +511,18 @@ yuck:
 		return -1;
 	}
 
-	/* Remove any existing entry */
+	/* Remove any existing entry, only http headers are left */
 	AST_LIST_LOCK(list);
-	AST_LIST_TRAVERSE_SAFE_BEGIN(list, cur, list) {
-		if (cur->key == new->key) {
-			AST_LIST_REMOVE_CURRENT(list);
-			ast_free(cur);
-			break;
+	if (new->key != CURLOPT_HTTPHEADER) {
+		AST_LIST_TRAVERSE_SAFE_BEGIN(list, cur, list) {
+			if (cur->key == new->key) {
+				AST_LIST_REMOVE_CURRENT(list);
+				ast_free(cur);
+				break;
+			}
 		}
+		AST_LIST_TRAVERSE_SAFE_END
 	}
-	AST_LIST_TRAVERSE_SAFE_END
 
 	/* Insert new entry */
 	ast_debug(1, "Inserting entry %p with key %d and value %p\n", new, new->key, new->value);
@@ -445,9 +546,17 @@ static int acf_curlopt_helper(struct ast_channel *chan, const char *cmd, char *d
 		return -1;
 	}
 
-	if (chan && (store = ast_channel_datastore_find(chan, &curl_info, NULL))) {
-		list[0] = store->data;
-		list[1] = &global_curl_info;
+	if (chan) {
+		/* If we have a channel, we want to read the options set there before
+		   falling back to the global settings */
+		ast_channel_lock(chan);
+		store = ast_channel_datastore_find(chan, &curl_info, NULL);
+		ast_channel_unlock(chan);
+
+		if (store) {
+			list[0] = store->data;
+			list[1] = &global_curl_info;
+		}
 	}
 
 	for (i = 0; i < 2; i++) {
@@ -573,8 +682,6 @@ static size_t WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *da
 	return realsize;
 }
 
-static const char * const global_useragent = "asterisk-libcurl-agent/1.0";
-
 static int curl_instance_init(void *data)
 {
 	CURL **curl = data;
@@ -585,7 +692,7 @@ static int curl_instance_init(void *data)
 	curl_easy_setopt(*curl, CURLOPT_NOSIGNAL, 1);
 	curl_easy_setopt(*curl, CURLOPT_TIMEOUT, 180);
 	curl_easy_setopt(*curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
-	curl_easy_setopt(*curl, CURLOPT_USERAGENT, global_useragent);
+	curl_easy_setopt(*curl, CURLOPT_USERAGENT, AST_CURL_USER_AGENT);
 
 	return 0;
 }
@@ -636,9 +743,14 @@ struct curl_args {
 static int acf_curl_helper(struct ast_channel *chan, struct curl_args *args)
 {
 	struct ast_str *escapebuf = ast_str_thread_get(&thread_escapebuf, 16);
-	int ret = -1;
+	int ret = 0;
+	long http_code = 0; /* read curl response */
+	size_t i;
+	struct ast_vector_int hasfailurecode = { NULL };
+	char *failurecodestrings,*found;
 	CURL **curl;
 	struct curl_settings *cur;
+	struct curl_slist *headers = NULL;
 	struct ast_datastore *store = NULL;
 	int hashcompat = 0;
 	AST_LIST_HEAD(global_curl_info, curl_settings) *list = NULL;
@@ -662,24 +774,44 @@ static int acf_curl_helper(struct ast_channel *chan, struct curl_args *args)
 		ast_autoservice_start(chan);
 	}
 
+	AST_VECTOR_INIT(&hasfailurecode, 0); /*Initialize vector*/
 	AST_LIST_LOCK(&global_curl_info);
 	AST_LIST_TRAVERSE(&global_curl_info, cur, list) {
 		if (cur->key == CURLOPT_SPECIAL_HASHCOMPAT) {
 			hashcompat = (long) cur->value;
+		} else if (cur->key == CURLOPT_HTTPHEADER) {
+			headers = curl_slist_append(headers, (char*) cur->value);
+		} else if (cur->key == CURLOPT_SPECIAL_FAILURE_CODE) {
+			failurecodestrings = (char*) cur->value;
+			while( (found = strsep(&failurecodestrings, ",")) != NULL) {
+				AST_VECTOR_APPEND(&hasfailurecode, atoi(found));
+			}
 		} else {
 			curl_easy_setopt(*curl, cur->key, cur->value);
 		}
 	}
 	AST_LIST_UNLOCK(&global_curl_info);
 
-	if (chan && (store = ast_channel_datastore_find(chan, &curl_info, NULL))) {
-		list = store->data;
-		AST_LIST_LOCK(list);
-		AST_LIST_TRAVERSE(list, cur, list) {
-			if (cur->key == CURLOPT_SPECIAL_HASHCOMPAT) {
-				hashcompat = (long) cur->value;
-			} else {
-				curl_easy_setopt(*curl, cur->key, cur->value);
+	if (chan) {
+		ast_channel_lock(chan);
+		store = ast_channel_datastore_find(chan, &curl_info, NULL);
+		ast_channel_unlock(chan);
+		if (store) {
+			list = store->data;
+			AST_LIST_LOCK(list);
+			AST_LIST_TRAVERSE(list, cur, list) {
+				if (cur->key == CURLOPT_SPECIAL_HASHCOMPAT) {
+					hashcompat = (long) cur->value;
+				} else if (cur->key == CURLOPT_HTTPHEADER) {
+					headers = curl_slist_append(headers, (char*) cur->value);
+				} else if (cur->key == CURLOPT_SPECIAL_FAILURE_CODE) {
+					failurecodestrings = (char*) cur->value;
+					while( (found = strsep(&failurecodestrings, ",")) != NULL) {
+						AST_VECTOR_APPEND(&hasfailurecode, atoi(found));
+					}
+				} else {
+					curl_easy_setopt(*curl, cur->key, cur->value);
+				}
 			}
 		}
 	}
@@ -692,6 +824,11 @@ static int acf_curl_helper(struct ast_channel *chan, struct curl_args *args)
 		curl_easy_setopt(*curl, CURLOPT_POSTFIELDS, args->postdata);
 	}
 
+	/* Always assign the headers - even when NULL - in case we had
+	 * custom headers the last time we used this shared cURL
+	 * instance */
+	curl_easy_setopt(*curl, CURLOPT_HTTPHEADER, headers);
+
 	/* Temporarily assign a buffer for curl to write errors to. */
 	curl_errbuf[0] = curl_errbuf[CURL_ERROR_SIZE] = '\0';
 	curl_easy_setopt(*curl, CURLOPT_ERRORBUFFER, curl_errbuf);
@@ -705,10 +842,25 @@ static int acf_curl_helper(struct ast_channel *chan, struct curl_args *args)
 	 * here, but the source allows it. See: "typecheck: allow NULL to unset
 	 * CURLOPT_ERRORBUFFER" (62bcf005f4678a93158358265ba905bace33b834). */
 	curl_easy_setopt(*curl, CURLOPT_ERRORBUFFER, (char*)NULL);
+	curl_easy_getinfo (*curl, CURLINFO_RESPONSE_CODE, &http_code);
+
+	for (i = 0; i < AST_VECTOR_SIZE(&hasfailurecode); ++i) {
+		if (http_code == AST_VECTOR_GET(&hasfailurecode,i)){
+			ast_log(LOG_NOTICE, "%s%sCURL '%s' returned response code (%ld).\n",
+				chan ? ast_channel_name(chan) : "",
+				chan ? ast_channel_name(chan) : ": ",
+				args->url,
+				http_code);
+			ret=-1;
+			break;
+		}
+	}
+	AST_VECTOR_FREE(&hasfailurecode); /* Release the vector*/
 
 	if (store) {
 		AST_LIST_UNLOCK(list);
 	}
+	curl_slist_free_all(headers);
 
 	if (args->postdata) {
 		curl_easy_setopt(*curl, CURLOPT_POST, 0);
@@ -740,7 +892,6 @@ static int acf_curl_helper(struct ast_channel *chan, struct curl_args *args)
 			ast_free(fields);
 			ast_free(values);
 		}
-		ret = 0;
 	}
 
 	if (chan) {
@@ -827,28 +978,6 @@ static struct ast_custom_function acf_curl = {
 
 static struct ast_custom_function acf_curlopt = {
 	.name = "CURLOPT",
-	.synopsis = "Set options for use with the CURL() function",
-	.syntax = "CURLOPT(<option>)",
-	.desc =
-"  cookie         - Send cookie with request [none]\n"
-"  conntimeout    - Number of seconds to wait for connection\n"
-"  dnstimeout     - Number of seconds to wait for DNS response\n"
-"  ftptext        - For FTP, force a text transfer (boolean)\n"
-"  ftptimeout     - For FTP, the server response timeout\n"
-"  header         - Retrieve header information (boolean)\n"
-"  httptimeout    - Number of seconds to wait for HTTP response\n"
-"  maxredirs      - Maximum number of redirects to follow\n"
-"  proxy          - Hostname or IP to use as a proxy\n"
-"  proxytype      - http, socks4, or socks5\n"
-"  proxyport      - port number of the proxy\n"
-"  proxyuserpwd   - A <user>:<pass> to use for authentication\n"
-"  referer        - Referer URL to use for the request\n"
-"  useragent      - UserAgent string to use\n"
-"  userpwd        - A <user>:<pass> to use for authentication\n"
-"  ssl_verifypeer - Whether to verify the peer certificate (boolean)\n"
-"  hashcompat     - Result data will be compatible for use with HASH()\n"
-"                 - if value is \"legacy\", will translate '+' to ' '\n"
-"",
 	.read = acf_curlopt_read,
 	.read2 = acf_curlopt_read2,
 	.write = acf_curlopt_write,
@@ -920,13 +1049,6 @@ static int load_module(void)
 {
 	int res;
 
-	if (!ast_module_check("res_curl.so")) {
-		if (ast_load_resource("res_curl.so") != AST_MODULE_LOAD_SUCCESS) {
-			ast_log(LOG_ERROR, "Cannot load res_curl, so func_curl cannot be loaded\n");
-			return AST_MODULE_LOAD_DECLINE;
-		}
-	}
-
 	res = ast_custom_function_register_escalating(&acf_curl, AST_CFE_WRITE);
 	res |= ast_custom_function_register(&acf_curlopt);
 
@@ -940,4 +1062,5 @@ AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Load external URL",
 	.load = load_module,
 	.unload = unload_module,
 	.load_pri = AST_MODPRI_REALTIME_DEPEND2,
+	.requires = "res_curl",
 );
